from typing import Any, Dict, List, Optional

from forestadmin.datasource_sqlalchemy.utils.type_converter import Converter as TypeConverter
from forestadmin.datasource_sqlalchemy.utils.type_converter import FilterOperator
from forestadmin.datasource_toolkit.interfaces.fields import (
    Column,
    FieldType,
    ManyToMany,
    ManyToOne,
    OneToMany,
    OneToOne,
    Operator,
    RelationAlias,
    Validation,
)
from forestadmin.datasource_toolkit.interfaces.models.collections import CollectionSchema
from sqlalchemy import Enum, Table
from sqlalchemy.orm import Mapper
from sqlalchemy.sql.schema import Column as SqlAlchemyColumn


class ColumnFactory:
    @staticmethod
    def _build_enum_values(column: SqlAlchemyColumn) -> Optional[List[str]]:
        if column.type.__class__ == Enum:  # type: ignore
            return column.type.enums  # type: ignore
        return None

    @staticmethod
    def _build_is_read_only(column: SqlAlchemyColumn) -> bool:
        return column.primary_key and column.autoincrement is not False  # type: ignore

    @classmethod
    def _build_validations(cls, column: SqlAlchemyColumn) -> List[Validation]:
        validations: List[Validation] = []
        if (
            not column.nullable
            and not cls._build_is_read_only(column)  # type: ignore
            and not column.default
            and not column.server_default  # type: ignore  # type: ignore
        ):
            validations.append(
                {
                    "operator": Operator.PRESENT,
                }
            )

        return validations

    @classmethod
    def build(cls, column: SqlAlchemyColumn) -> Column:
        column_type = TypeConverter.convert(column.type)  # type: ignore
        return {
            "column_type": column_type,
            "is_primary_key": column.primary_key,  # type: ignore
            "is_read_only": cls._build_is_read_only(column),
            "default_value": column.default,  # type: ignore
            "is_sortable": None,
            "validations": cls._build_validations(column),
            "filter_operators": FilterOperator.get_for_type(column_type),
            "enum_values": cls._build_enum_values(column),
            "type": FieldType.COLUMN,
        }


def is_composite_fk(relation: Any):
    return len(relation.remote_side) > 1 or len(relation.local_columns) > 1


class CollectionFactory:
    @staticmethod
    def _build_one_to_many(relation: Any) -> Optional[OneToMany]:
        if is_composite_fk(relation):
            return None
        return {
            "foreign_collection": relation.target.name,
            "origin_key": list(relation.remote_side)[0].name,
            "origin_key_target": list(relation.local_columns)[0].name,
            "type": FieldType.ONE_TO_MANY,
        }

    @staticmethod
    def _build_many_to_one(relation: Any) -> Optional[ManyToOne]:
        if is_composite_fk(relation):
            return None
        return {
            "foreign_collection": relation.target.name,
            "foreign_key": list(relation.local_columns)[0].name,
            "foreign_key_target": list(relation.remote_side)[0].name,
            "type": FieldType.MANY_TO_ONE,
        }

    @staticmethod
    def _build_one_to_one(relation: Any) -> Optional[OneToOne]:
        if is_composite_fk(relation):
            return None
        return {
            "foreign_collection": relation.target.name,
            "origin_key": list(relation.remote_side)[0].name,
            "origin_key_target": list(relation.local_columns)[0].name,
            "type": FieldType.ONE_TO_ONE,
        }

    @staticmethod
    def _build_many_to_many(model: Table, relation: Any) -> Optional[ManyToMany]:
        kwargs: Dict[str, str] = {}
        for column in relation.remote_side:
            if len(column.foreign_keys) > 1:  # composite fk
                return None
            fk = list(column.foreign_keys)[0]
            if fk.column.table.name == model.name:
                kwargs["origin_key_target"] = fk.column.name
                kwargs["origin_key"] = column.name
            else:
                kwargs["foreign_key_target"] = fk.column.name
                kwargs["foreign_key"] = column.name
                kwargs["foreign_collection"] = fk.column.table.name
        return ManyToMany(
            **{
                "through_collection": relation.secondary.name,  # type: ignore
                "type": FieldType.MANY_TO_MANY,  # type: ignore
                "foreign_relation": None,  # type: ignore
                **kwargs,
            }
        )

    @classmethod
    def build(cls, model: Table, mapper: Optional[Mapper] = None) -> CollectionSchema:
        fields = {}
        for column in model.columns:  # type: ignore
            fields[column.name] = ColumnFactory.build(column)  # type: ignore

        if mapper:
            for relationship in mapper.relationships:  # type: ignore
                relation: Optional[RelationAlias] = None
                if not relationship.back_populates:  # type: ignore
                    # one to many
                    relation = cls._build_one_to_many(relationship)
                else:
                    if relationship.uselist is False:  # type: ignore
                        if list(relationship.local_columns)[0].foreign_keys:  # type: ignore
                            relation = cls._build_many_to_one(relationship)
                        else:
                            relation = cls._build_one_to_one(relationship)
                    elif relationship.secondary is not None:  # type: ignore
                        relation = cls._build_many_to_many(model, relationship)
                    else:
                        relation = cls._build_one_to_many(relationship)

                    if relation:
                        fields[relationship.key] = relation  # type: ignore

        return {"actions": {}, "fields": fields, "searchable": False, "segments": []}



import abc
from typing import Any, Callable, List, Optional, Tuple

from forestadmin.datasource_sqlalchemy.utils.relationships import Relationships
from forestadmin.datasource_toolkit.collections import Collection
from forestadmin.datasource_toolkit.interfaces.query.projections import Projection
from forestadmin.datasource_toolkit.interfaces.records import RecordsDataAlias
from sqlalchemy import Table
from sqlalchemy import column as SqlAlchemyColumn
from sqlalchemy.orm import Mapper
from typing_extensions import Self


class BaseSqlAlchemyCollectionFactory(abc.ABC):
    @abc.abstractmethod
    def init_instance(self, data: RecordsDataAlias) -> "BaseSqlAlchemyCollection":
        pass


class BaseSqlAlchemyCollection(Collection, abc.ABC):
    @abc.abstractmethod
    def get_column(self, name: str) -> Self:
        pass

    @abc.abstractmethod
    def _get_related_column(
        self, projection: Projection, level: int = 0
    ) -> Tuple[List[SqlAlchemyColumn], Relationships]:
        pass

    @abc.abstractmethod
    def get_columns(self, projection: Projection, level: int = 0) -> Tuple[List[SqlAlchemyColumn], Relationships]:
        pass

    @abc.abstractproperty
    def table(self) -> Table:  # type: ignore
        pass

    @abc.abstractproperty
    def mapper(self) -> Optional[Mapper]:
        pass

    @abc.abstractproperty
    def model(self) -> Optional[Callable[[Any], Any]]:
        pass

    @abc.abstractproperty
    def factory(self) -> BaseSqlAlchemyCollectionFactory:  # type: ignore
        pass



from collections import defaultdict
from typing import Callable, List, Tuple

from forestadmin.datasource_sqlalchemy.interfaces import BaseSqlAlchemyCollection
from forestadmin.datasource_sqlalchemy.utils.relationships import Relationships, merge_relationships
from forestadmin.datasource_toolkit.exceptions import DatasourceToolkitException
from forestadmin.datasource_toolkit.interfaces.query.aggregation import Aggregation, Aggregator, DateOperation
from forestadmin.datasource_toolkit.interfaces.query.projections import Projection
from sqlalchemy import DATE, cast
from sqlalchemy import column as SqlAlchemyColumn
from sqlalchemy import func, text
from sqlalchemy.engine import Dialect


class AggregationFactoryException(DatasourceToolkitException):
    pass


class AggregationFactory:

    LABEL = "__aggregate__"
    GROUP_LABEL = "__grouped__"

    MAPPING = {
        Aggregator.COUNT: func.count,
        Aggregator.MIN: func.min,
        Aggregator.MAX: func.max,
        Aggregator.AVG: func.avg,
        Aggregator.SUM: func.sum,
    }

    @staticmethod
    def get_func(aggregation: Aggregation) -> Callable[[str], SqlAlchemyColumn]:
        try:
            return AggregationFactory.MAPPING[aggregation.operation]
        except KeyError:
            raise AggregationFactoryException(f'Unknown aggregator "{aggregation.operation}"')

    @classmethod
    def get_group_field_name(cls, field_name: str):
        return f"{field_name}{cls.GROUP_LABEL}"

    @classmethod
    def get_field_from_group_field_name(cls, group_field_name: str):
        return group_field_name[: -len(cls.GROUP_LABEL)]

    @staticmethod
    def get_field(collection: BaseSqlAlchemyCollection, aggregation: Aggregation) -> Tuple[str, Relationships]:
        field = "*"
        relationships: Relationships = defaultdict(list)
        if aggregation.field:
            fields, relationships = collection.get_columns(Projection(aggregation.field))
            field = fields[0]
        return field, relationships

    @classmethod
    def build_column(cls, collection: BaseSqlAlchemyCollection, aggregation: Aggregation) -> SqlAlchemyColumn:
        func = cls.get_func(aggregation)
        field, relationships = cls.get_field(collection, aggregation)
        return func(field).label(cls.LABEL), relationships

    @classmethod
    def build_group(
        cls,
        dialect: Dialect,
        collection: BaseSqlAlchemyCollection,
        aggregation: Aggregation,
    ) -> Tuple[List[SqlAlchemyColumn], Relationships]:
        groups: List[SqlAlchemyColumn] = []
        group_relationships: Relationships = defaultdict(list)
        for group in aggregation.groups:
            columns, relationships = collection.get_columns(Projection(group["field"]))
            operation = group.get("operation")
            if operation:
                column = DateAggregation.build(dialect, columns[0], operation)
            else:
                column = columns[0]
            groups.append(column.label(cls.get_group_field_name(group["field"])))
            group_relationships = merge_relationships(relationships, group_relationships)
        return groups, group_relationships


class DateAggregation:
    @staticmethod
    def build_postgres(column: SqlAlchemyColumn, operation: DateOperation) -> SqlAlchemyColumn:
        return func.date_trunc(operation.value.lower(), column)

    @staticmethod
    def build_sqllite(column: SqlAlchemyColumn, operation: DateOperation) -> SqlAlchemyColumn:
        if operation == DateOperation.WEEK:
            return func.DATE(column, "weekday 1", "-7 days")
        elif operation == DateOperation.YEAR:
            format = "%Y-01-01"
        elif operation == DateOperation.MONTH:
            format = "%Y-%m-01"
        elif operation == DateOperation.DAY:
            format = "%Y-%m-%d"
        else:
            raise AggregationFactoryException()
        return func.strftime(format, column)

    @staticmethod
    def build_mysql(column: SqlAlchemyColumn, operation: DateOperation) -> SqlAlchemyColumn:
        format = "%Y-%m-%d"
        if operation == DateOperation.YEAR:
            format = "%Y-01-01"
        elif operation == DateOperation.MONTH:
            format = "%Y-%m-01"
        elif operation == DateOperation.WEEK:
            return cast(func.date_sub(column, text(f"INTERVAL(WEEKDAY({column})) DAY")), DATE)
        elif operation == DateOperation.DAY:
            format = "%Y-%m-%d"
        else:
            raise AggregationFactoryException()
        return func.date_format(column, format)

    @staticmethod
    def build_mssql(column: SqlAlchemyColumn, operation: DateOperation) -> SqlAlchemyColumn:
        if operation == DateOperation.YEAR:
            return func.datefromparts(func.extract("year", column), "01", "01")
        elif operation == DateOperation.MONTH:
            return func.datefromparts(func.extract("year", column), func.extract("month", column), "01")
        elif operation == DateOperation.WEEK:
            return cast(func.dateadd(text("day"), -func.extract("dw", column) + 2, column), DATE)
        elif operation == DateOperation.DAY:
            return func.datefromparts(
                func.extract("year", column),
                func.extract("month", column),
                func.extract("day", column),
            )

    @classmethod
    def build(cls, dialect: Dialect, column: SqlAlchemyColumn, operation: DateOperation) -> SqlAlchemyColumn:
        if dialect.name == "sqlite":
            return cls.build_sqllite(column, operation)
        elif dialect.name in ["mysql", "mariadb"]:
            return cls.build_mysql(column, operation)
        elif dialect.name == "postgresql":
            return cls.build_postgres(column, operation)
        elif dialect.name == "mssql":
            return cls.build_mssql(column, operation)
        else:
            raise AggregationFactoryException(f"The dialect {dialect.name} is not handled")



import sys
from collections import defaultdict

if sys.version_info >= (3, 8):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict

from typing import Any, Dict, List, Optional, Tuple, cast

from forestadmin.datasource_sqlalchemy.interfaces import BaseSqlAlchemyCollection
from forestadmin.datasource_sqlalchemy.utils.aggregation import AggregationFactory
from forestadmin.datasource_sqlalchemy.utils.relationships import Relationships, merge_relationships
from forestadmin.datasource_sqlalchemy.utils.type_converter import FilterOperator
from forestadmin.datasource_toolkit.exceptions import DatasourceToolkitException
from forestadmin.datasource_toolkit.interfaces.query.aggregation import Aggregation
from forestadmin.datasource_toolkit.interfaces.query.condition_tree.nodes.base import ConditionTree
from forestadmin.datasource_toolkit.interfaces.query.condition_tree.nodes.branch import Aggregator, ConditionTreeBranch
from forestadmin.datasource_toolkit.interfaces.query.condition_tree.nodes.leaf import ConditionTreeLeaf
from forestadmin.datasource_toolkit.interfaces.query.filter.paginated import PaginatedFilter
from forestadmin.datasource_toolkit.interfaces.query.filter.unpaginated import Filter
from forestadmin.datasource_toolkit.interfaces.query.projections import Projection
from forestadmin.datasource_toolkit.interfaces.query.sort import PlainSortClause
from forestadmin.datasource_toolkit.interfaces.records import RecordsDataAlias
from sqlalchemy import and_
from sqlalchemy import column as SqlAlchemyColumn
from sqlalchemy import delete, or_, select, update
from sqlalchemy.engine import Dialect
from sqlalchemy.sql.elements import BooleanClauseList, UnaryExpression


class ConditionTreeFactoryException(DatasourceToolkitException):
    pass


class ConditionTreeFactory:

    AGGREGATORS = {Aggregator.AND: and_, Aggregator.OR: or_}

    @classmethod
    def _build_leaf_condition(cls, collection: BaseSqlAlchemyCollection, leaf: ConditionTreeLeaf) -> Tuple[Any, Any]:
        projection = leaf.projection
        columns, relationships = collection.get_columns(projection)
        operator = FilterOperator.get_operator(columns, leaf.operator)
        return operator(leaf.value), relationships

    @classmethod
    def _get_aggregator(cls, aggregator: Aggregator):
        try:
            return cls.AGGREGATORS[aggregator]
        except KeyError:
            raise DatasourceToolkitException(f"Unable to handle the aggregator {aggregator}")

    @classmethod
    def _build_branch_condition(
        cls, collection: BaseSqlAlchemyCollection, branch: ConditionTreeBranch
    ) -> Tuple[Any, Any]:
        relationships: Dict[int, List[SqlAlchemyColumn]] = defaultdict(list)
        aggregator = cls._get_aggregator(branch.aggregator)
        clauses: List[Any] = []
        for condition in branch.conditions:
            clause, leaf_relationships = cls._build(collection, condition)
            clauses.append(clause)
            relationships = merge_relationships(relationships, leaf_relationships)
        return aggregator(*clauses), relationships

    @classmethod
    def _build(cls, collection: BaseSqlAlchemyCollection, condition_tree: ConditionTree):
        if isinstance(condition_tree, ConditionTreeLeaf):
            return cls._build_leaf_condition(collection, condition_tree)
        elif isinstance(condition_tree, ConditionTreeBranch):
            return cls._build_branch_condition(collection, condition_tree)
        raise ConditionTreeFactoryException(f"Unable to handle the type {condition_tree.__class__}")

    @classmethod
    def build(cls, collection: BaseSqlAlchemyCollection, condition_tree: ConditionTree):
        return cls._build(collection, condition_tree)


class FilterOptions(TypedDict):
    clauses: Optional[BooleanClauseList]
    relationships: Relationships


class PaginatedFilterOptions(FilterOptions):
    order_by: Optional[List[UnaryExpression]]


class FilterFactory:
    @classmethod
    def build(cls, collection: BaseSqlAlchemyCollection, filter: Optional[Filter]) -> FilterOptions:
        res: FilterOptions = {
            "relationships": defaultdict(list),
            "clauses": None,
        }
        if filter and filter.condition_tree:
            res["clauses"], clauses_relationships = ConditionTreeFactory.build(collection, filter.condition_tree)
            res["relationships"] = merge_relationships(res["relationships"], clauses_relationships)
        return res


class PaginatedFilterFactory:
    @staticmethod
    def get_order_by(collection: BaseSqlAlchemyCollection, filter: PaginatedFilter) -> Tuple[List[Any], Relationships]:
        relationships: Relationships = defaultdict(list)
        order_clauses: List[Any] = []
        for sort in cast(List[PlainSortClause], filter.sort or []):
            columns, nested_relationships = collection.get_columns(Projection(sort["field"]))
            relationships = merge_relationships(relationships, nested_relationships)
            if sort["ascending"]:
                order_clauses.append(columns[0].asc())  # type: ignore
            else:
                order_clauses.append(columns[0].desc())  # type: ignore
        return order_clauses, relationships

    @classmethod
    def build(cls, collection: BaseSqlAlchemyCollection, filter: PaginatedFilter) -> PaginatedFilterOptions:
        filter_ = FilterFactory.build(collection, filter.to_base_filter())
        res: PaginatedFilterOptions = {
            "relationships": filter_["relationships"],
            "clauses": filter_["clauses"],
            "order_by": None,
        }

        if filter.sort:
            res["order_by"], order_by_relationships = cls.get_order_by(collection, filter)
            res["relationships"] = merge_relationships(res["relationships"], order_by_relationships)
        return res


class QueryFactoryException(DatasourceToolkitException):
    pass


class QueryFactory:
    @classmethod
    def _build_list(
        cls,
        collection: BaseSqlAlchemyCollection,
        filter: Optional[PaginatedFilter],
        columns: List[SqlAlchemyColumn],
        relationships: Optional[Relationships] = None,
    ):
        if not relationships:
            relationships = defaultdict(list)
        query: Any = select(columns).select_from(collection.mapper)
        if filter:
            options = PaginatedFilterFactory.build(collection, filter)
            relationships = merge_relationships(relationships, options.get("relationships", {}))

            if options.get("clauses") is not None:
                query = query.where(options["clauses"])
            if options.get("order_by"):
                query = query.order_by(*options.get("order_by"))

            if filter.page:
                query = query.limit(filter.page.limit).offset(filter.page.skip)

        for level in sorted(relationships.keys()):
            query = query.join(*relationships[level], isouter=True)

        return query

    @classmethod
    def build_list(
        cls,
        collection: BaseSqlAlchemyCollection,
        filter: PaginatedFilter,
        projection: Projection,
    ):
        if collection.mapper:
            columns, relationships = collection.get_columns(projection)
            return cls._build_list(collection, filter, columns, relationships)  # type: ignore
        raise QueryFactoryException("Unable to request a collection without mapper")

    @classmethod
    def build_aggregate(
        cls,
        dialect: Dialect,
        collection: BaseSqlAlchemyCollection,
        filter: Optional[Filter],
        aggregation: Aggregation,
        limit: Optional[int],
    ):
        if collection.mapper:
            column, relationships = AggregationFactory.build_column(collection, aggregation)
            _filter = None
            if filter is not None:
                _filter = PaginatedFilter.from_base_filter(filter)
            groups, group_relationships = AggregationFactory.build_group(dialect, collection, aggregation)
            query = cls._build_list(
                collection,
                _filter,
                [column, *groups],
                merge_relationships(relationships, group_relationships),
            )
            if groups:
                query = query.group_by(*groups)

            if dialect.name == "postgresql":
                query = query.order_by(column.desc().nulls_last())
            else:
                query = query.order_by(column.desc())

            if limit:
                query = query.limit(limit)
            return query

    @staticmethod
    def update(
        collection: BaseSqlAlchemyCollection,
        filter: Optional[Filter],
        patch: RecordsDataAlias,
    ):
        if collection.model:
            options: FilterOptions = FilterFactory.build(collection, filter)
            query = update(collection.model).values(**patch)
            if options["clauses"] is not None:
                query = query.where(options["clauses"])
            return query.execution_options(synchronize_session="fetch")
        raise QueryFactoryException("Unable to request a collection without mapper")

    @staticmethod
    def create(collection: BaseSqlAlchemyCollection, data: List[RecordsDataAlias]):
        if collection.model:
            instances: List[Any] = []
            for record in data:
                instances.append(collection.factory.init_instance(record))
            return instances
        raise QueryFactoryException("Unable to request a collection without mapper")

    @staticmethod
    def delete(collection: BaseSqlAlchemyCollection, filter: Optional[Filter]):
        if collection.model:
            options: FilterOptions = FilterFactory.build(collection, filter)
            query = delete(collection.model)
            if options["clauses"] is not None:
                query = query.where(options["clauses"])
            return query.execution_options(synchronize_session="fetch")
        raise QueryFactoryException("Unable to request a collection without mapper")



from typing import Any, Callable, Dict, Optional, Set, Type

from forestadmin.datasource_toolkit.exceptions import DatasourceToolkitException
from forestadmin.datasource_toolkit.interfaces.fields import ColumnAlias, Operator, PrimitiveType
from sqlalchemy import ARRAY  # type: ignore
from sqlalchemy import column as SqlAlchemyColumn  # type: ignore
from sqlalchemy import func, not_, or_  # type: ignore
from sqlalchemy import types as sqltypes
from sqlalchemy.dialects.postgresql import UUID


class ConverterException(DatasourceToolkitException):
    pass


class Converter:

    TYPES: Dict[Type[sqltypes.TypeEngine], PrimitiveType] = {
        sqltypes.BigInteger: PrimitiveType.NUMBER,
        sqltypes.Boolean: PrimitiveType.BOOLEAN,
        sqltypes.CHAR: PrimitiveType.STRING,
        sqltypes.Date: PrimitiveType.DATE_ONLY,
        sqltypes.DateTime: PrimitiveType.DATE,
        sqltypes.DECIMAL: PrimitiveType.NUMBER,
        sqltypes.Enum: PrimitiveType.ENUM,
        sqltypes.Float: PrimitiveType.NUMBER,
        sqltypes.Integer: PrimitiveType.NUMBER,
        sqltypes.JSON: PrimitiveType.JSON,
        sqltypes.Numeric: PrimitiveType.NUMBER,
        sqltypes.REAL: PrimitiveType.NUMBER,
        sqltypes.SmallInteger: PrimitiveType.NUMBER,
        sqltypes.String: PrimitiveType.STRING,
        sqltypes.Text: PrimitiveType.STRING,
        sqltypes.Time: PrimitiveType.TIME_ONLY,
        sqltypes.Unicode: PrimitiveType.STRING,
        sqltypes.UnicodeText: PrimitiveType.STRING,
        UUID: PrimitiveType.UUID,
    }

    @classmethod
    def convert(cls, _type: sqltypes.TypeEngine) -> ColumnAlias:
        type_class = _type.__class__
        if type_class == ARRAY:
            return [cls.convert(_type.item_type)]  # type: ignore
        try:
            return cls.TYPES[_type.__class__]
        except KeyError:
            raise ConverterException(f'Type "{_type.__class__}" is unknown')


class FilterOperator:

    COMMON_OPERATORS: Set[Operator] = {
        Operator.BLANK,
        Operator.EQUAL,
        Operator.MISSING,
        Operator.NOT_EQUAL,
        Operator.PRESENT,
    }

    OPERATORS = {
        Operator.EQUAL: "_equal_operator",
        Operator.NOT_EQUAL: "_not_equal_operator",
        Operator.BLANK: "_blank_operator",
        Operator.CONTAINS: "_contains_operator",
        Operator.NOT_CONTAINS: "_not_contains_operator",
        Operator.STARTS_WITH: "_starts_with_operator",
        Operator.ENDS_WITH: "_ends_with_operator",
        Operator.GREATER_THAN: "_greater_than_operator",
        Operator.LESS_THAN: "_less_than_operator",
        Operator.MISSING: "_missing_operator",
        Operator.PRESENT: "_present_operator",
        Operator.IN: "_in_operator",
        Operator.NOT_IN: "_not_in_operator",
        Operator.INCLUDES_ALL: "_includes_all",
    }

    @staticmethod
    def _equal_operator(column: SqlAlchemyColumn):
        return column.__eq__

    @staticmethod
    def _not_equal_operator(column: SqlAlchemyColumn):
        return column.__ne__

    @staticmethod
    def _blank_operator(column: SqlAlchemyColumn):
        def wrapped(_: str):
            return or_([column.is_(None), column.__eq__("")])

        return wrapped

    @staticmethod
    def _contains_operator(column: SqlAlchemyColumn):
        def wrapped(value: str):
            return func.lower(column).contains(value.lower())

        return wrapped

    @classmethod
    def _not_contains_operator(cls, column: SqlAlchemyColumn):
        def wrapped(value: str) -> Any:
            return not_(cls._contains_operator(column)(value))  # type: ignore

        return wrapped

    @staticmethod
    def _starts_with_operator(column: SqlAlchemyColumn):
        return column.startswith

    @staticmethod
    def _ends_with_operator(column: SqlAlchemyColumn):
        return column.endswith

    @staticmethod
    def _greater_than_operator(column: SqlAlchemyColumn):
        return column.__gt__

    @staticmethod
    def _less_than_operator(column: SqlAlchemyColumn):
        return column.__lt__

    @staticmethod
    def _in_operator(column: SqlAlchemyColumn):
        return column.in_

    @staticmethod
    def _not_in_operator(column: SqlAlchemyColumn):
        return column.not_in

    @staticmethod
    def _missing_operator(column: SqlAlchemyColumn):
        def wrapped(_: str):
            return column.is_(None)

        return wrapped

    @staticmethod
    def _present_operator(column: SqlAlchemyColumn):
        def wrapped(_: str):
            return column.is_not(None)

        return wrapped

    @staticmethod
    def _includes_all(column: SqlAlchemyColumn):
        return column.__eq__

    @classmethod
    def get_operator(cls, columns: SqlAlchemyColumn, operator: Operator) -> Callable[[Optional[str]], Any]:
        try:
            meth = cls.OPERATORS[operator]
        except KeyError:
            raise ConverterException(f"Unable to handle the operator {operator}")
        else:
            return getattr(cls, meth)(columns[0])

    @classmethod
    def get_for_type(cls, _type: ColumnAlias) -> set[Operator]:

        operators: Set[Operator] = set()
        if isinstance(_type, list):
            operators = {
                *cls.COMMON_OPERATORS,
                Operator.IN,
                Operator.INCLUDES_ALL,
                Operator.NOT_IN,
            }
        elif _type == PrimitiveType.BOOLEAN:
            operators = cls.COMMON_OPERATORS
        elif _type == PrimitiveType.UUID:
            operators = {
                *cls.COMMON_OPERATORS,
                Operator.CONTAINS,
                Operator.ENDS_WITH,
                Operator.LIKE,
                Operator.STARTS_WITH,
            }
        elif _type == PrimitiveType.NUMBER:
            operators = {
                *cls.COMMON_OPERATORS,
                Operator.GREATER_THAN,
                Operator.LESS_THAN,
                Operator.IN,
                Operator.NOT_IN,
            }
        elif _type == PrimitiveType.STRING:
            operators = {
                *cls.COMMON_OPERATORS,
                Operator.CONTAINS,
                Operator.ENDS_WITH,
                Operator.IN,
                Operator.LIKE,
                Operator.LONGER_THAN,
                Operator.NOT_CONTAINS,
                Operator.NOT_IN,
                Operator.SHORTER_THAN,
                Operator.STARTS_WITH,
            }
        elif _type in [
            PrimitiveType.DATE,
            PrimitiveType.DATE_ONLY,
            PrimitiveType.TIME_ONLY,
        ]:
            operators = {
                *cls.COMMON_OPERATORS,
                Operator.GREATER_THAN,
                Operator.LESS_THAN,
            }
        elif _type == PrimitiveType.ENUM:
            operators = {*cls.COMMON_OPERATORS, Operator.IN, Operator.NOT_IN}
        elif _type == PrimitiveType.JSON:
            operators = cls.COMMON_OPERATORS

        return operators



import sys

if sys.version_info >= (3, 9):
    import zoneinfo
else:
    from backports import zoneinfo

from datetime import datetime
from typing import Any, Dict, List, Tuple, cast

from forestadmin.datasource_sqlalchemy.interfaces import BaseSqlAlchemyCollection
from forestadmin.datasource_sqlalchemy.utils.aggregation import AggregationFactory
from forestadmin.datasource_toolkit.interfaces.query.aggregation import AggregateResult
from forestadmin.datasource_toolkit.interfaces.query.projections import Projection
from forestadmin.datasource_toolkit.interfaces.records import RecordsDataAlias


def instances_to_records(collection: BaseSqlAlchemyCollection, instances: List[Any]) -> List[RecordsDataAlias]:
    records: List[RecordsDataAlias] = []
    for instance in instances:
        record: RecordsDataAlias = {}
        for column in collection.mapper.c:  # type: ignore
            record[column.name] = getattr(instance, column.name)
        records.append(record)
    return records


def _cast_value(value: Any, timezone: zoneinfo.ZoneInfo) -> Any:
    if isinstance(value, datetime):
        value = value.replace(tzinfo=zoneinfo.ZoneInfo("UTC"))
    return value


def projections_to_records(
    projection: Projection, items: List[Tuple[Any]], timezone: zoneinfo.ZoneInfo
) -> List[RecordsDataAlias]:
    records: List[RecordsDataAlias] = []
    for item in items:
        result = dict(zip(cast(List[str], projection), item))
        record: RecordsDataAlias = dict([(field_name, result[field_name]) for field_name in projection.columns])
        for field_name, sub_fields in projection.relations.items():
            res: Dict[str, Any] = {}
            for sub_field in cast(List[str], sub_fields):
                value = result[f"{field_name}:{sub_field}"]
                if ":" in sub_field:
                    key, *sub_field = sub_field.split(":")
                    if value is not None:
                        for v in cast(List[str], reversed(sub_field)):
                            value = {f"{v}": _cast_value(value, timezone)}
                    res[key] = value
                else:
                    res[sub_field] = value
            else:
                if any(res.values()):
                    record[field_name] = res
                else:
                    record[field_name] = None
        for key in record:
            record[key] = _cast_value(record[key], timezone)
        records.append(record)
    return records


def aggregations_to_records(items: List[Dict[str, Any]]):
    records: List[AggregateResult] = []
    for item in items:
        result = AggregateResult(value=item[AggregationFactory.LABEL], group={})
        for key in item.keys():
            if AggregationFactory.GROUP_LABEL in key:
                result["group"][AggregationFactory.get_field_from_group_field_name(key)] = item[key]
        records.append(result)
    return records



from typing import DefaultDict, List

from sqlalchemy import column as SqlAlchemyColumn

Relationships = DefaultDict[int, List[SqlAlchemyColumn]]


def merge_relationships(r1: Relationships, r2: Relationships) -> Relationships:
    for level, nested_relationship in r2.items():
        r1[level].extend(nested_relationship)
        r1[level] = list(set(r1[level]))
    return r1



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['forestadmin',
 'forestadmin.datasource_sqlalchemy',
 'forestadmin.datasource_sqlalchemy.utils']

package_data = \
{'': ['*']}

install_requires = \
['forestadmin-datasource-toolkit>=0.1.0-beta.4,<0.2.0',
 'typing-extensions>=4.2,<5.0',
 'tzdata>=2022.6,<2023.0']

extras_require = \
{':python_version < "3.9"': ['backports.zoneinfo[tzdata]>=0.2.1,<0.3.0']}

setup_kwargs = {
    'name': 'forestadmin-datasource-sqlalchemy',
    'version': '0.1.0b4',
    'description': '',
    'long_description': '',
    'author': 'Valentin Monté',
    'author_email': 'valentinm@forestadmin.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import zoneinfo
from collections import defaultdict
from datetime import datetime
from typing import Any, Awaitable, Callable, Dict, List, Optional, Tuple, Union, cast

from forestadmin.datasource_sqlalchemy.interfaces import BaseSqlAlchemyCollection, BaseSqlAlchemyCollectionFactory
from forestadmin.datasource_sqlalchemy.utils.model_converter import CollectionFactory
from forestadmin.datasource_sqlalchemy.utils.query_factory import QueryFactory
from forestadmin.datasource_sqlalchemy.utils.record_serializer import (
    aggregations_to_records,
    instances_to_records,
    projections_to_records,
)
from forestadmin.datasource_sqlalchemy.utils.relationships import Relationships, merge_relationships
from forestadmin.datasource_toolkit.datasources import Datasource, DatasourceException
from forestadmin.datasource_toolkit.interfaces.actions import ActionField, ActionResult
from forestadmin.datasource_toolkit.interfaces.fields import (
    FieldType,
    ManyToMany,
    ManyToOne,
    PrimitiveType,
    RelationAlias,
)
from forestadmin.datasource_toolkit.interfaces.query.aggregation import AggregateResult, Aggregation
from forestadmin.datasource_toolkit.interfaces.query.condition_tree.nodes.base import ConditionTree
from forestadmin.datasource_toolkit.interfaces.query.condition_tree.nodes.leaf import ConditionTreeLeaf
from forestadmin.datasource_toolkit.interfaces.query.filter.paginated import PaginatedFilter
from forestadmin.datasource_toolkit.interfaces.query.filter.unpaginated import Filter
from forestadmin.datasource_toolkit.interfaces.query.projections import Projection
from forestadmin.datasource_toolkit.interfaces.records import RecordsDataAlias
from forestadmin.datasource_toolkit.validations.type_getter import TypeGetter
from sqlalchemy import Table
from sqlalchemy import column as SqlAlchemyColumn
from sqlalchemy.engine import Dialect
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm import Mapper, RelationshipProperty, sessionmaker


class SqlAlchemyCollectionException(DatasourceException):
    pass


def handle_sqlalchemy_error(fn: Callable[..., Awaitable[Any]]):
    async def wrapped(self: Any, *args: Any, **kwargs: Any) -> Any:
        try:
            return await fn(self, *args, **kwargs)
        except SQLAlchemyError as e:
            raise SqlAlchemyCollectionException(str(e))

    return wrapped


class SqlAlchemyCollectionFactory(BaseSqlAlchemyCollectionFactory):
    def __init__(self, collection: "SqlAlchemyCollection"):
        self._collection = collection

    @property
    def collection(self) -> "SqlAlchemyCollection":
        return self._collection

    def _unknown_fields(self, data: RecordsDataAlias) -> List[str]:
        fields: List[str] = []
        for name in data.keys():
            try:
                self.collection.get_column(name)
            except SqlAlchemyCollectionException:
                fields.append(name)
        return fields

    def init_instance(self, data: RecordsDataAlias) -> "SqlAlchemyCollection":
        if self.collection.model:
            try:
                return self.collection.model(**data)  # type: ignore
            except TypeError as e:
                unknown_fields = self._unknown_fields(data)
                if unknown_fields:
                    raise SqlAlchemyCollectionException(
                        f'Unknow fields "{unknown_fields}" for the model "{self.collection.model.__class__.__name__}"'
                    )
                raise e
        raise


class SqlAlchemyCollection(BaseSqlAlchemyCollection):
    def __init__(
        self,
        name: str,
        datasource: "SqlAlchemyDatasource",
        table: Table,
        mapper: Optional[Mapper] = None,
    ):
        super(SqlAlchemyCollection, self).__init__(name, datasource)
        self._table = table
        self._mapper = mapper
        self._name = name
        self._factory = SqlAlchemyCollectionFactory(self)
        schema = CollectionFactory.build(self.table, self.mapper)
        self.add_fields(schema["fields"])

    @property
    def table(self) -> Table:
        return self._table

    @property
    def mapper(self) -> Optional[Mapper]:
        return self._mapper

    @property
    def model(self) -> Optional[Callable[[Any], Any]]:
        if self.mapper:
            return self.mapper.class_
        return None

    @property
    def factory(self) -> SqlAlchemyCollectionFactory:
        return self._factory

    def get_column(self, name: str) -> SqlAlchemyColumn:
        try:
            return self.mapper.columns[name]  # type: ignore
        except KeyError:
            raise SqlAlchemyCollectionException(f"Unkown field '{name}' for the collection '{self.name}'")

    def _get_relationship(self, name: str) -> RelationshipProperty:
        try:
            return self.mapper.relationships[name]  # type: ignore
        except KeyError:
            raise SqlAlchemyCollectionException(f"Unkown relationship '{name}' for the collection '{self.name}'")

    def get_columns(self, projection: Projection, level: int = 0) -> Tuple[List[SqlAlchemyColumn], Relationships]:
        relationships: Relationships = defaultdict(list)
        columns: List[SqlAlchemyColumn] = [self.get_column(column) for column in projection.columns]
        nested_columns, nested_relationships = self._get_related_column(projection, level)
        columns.extend(nested_columns)
        relationships = merge_relationships(relationships, nested_relationships)
        return columns, relationships

    def _get_related_column(
        self, projection: Projection, level: int = 0
    ) -> Tuple[List[SqlAlchemyColumn], Relationships]:
        relationships: Dict[int, List[SqlAlchemyColumn]] = defaultdict(list)
        columns: List[SqlAlchemyColumn] = []
        for related_field_name, related_projection in projection.relations.items():
            relationships[level].append(self._get_relationship(related_field_name).class_attribute)  # type: ignore
            related_field: RelationAlias = cast(RelationAlias, self.get_field(related_field_name))
            related_collection = self.datasource.get_collection(related_field["foreign_collection"])
            nested_columns, nested_relationships = related_collection.get_columns(related_projection, level + 1)
            columns.extend(nested_columns)
            merge_relationships(relationships, nested_relationships)
        return columns, relationships

    def _normalize_projection(self, projection: Projection):
        # needed to be compliant with the orm result orm
        normalized_projection = projection.columns
        for parent_field, child_fields in projection.relations.items():
            for field in cast(List[str], child_fields):
                normalized_projection.append(f"{parent_field}:{field}")
        return Projection(*normalized_projection)

    async def execute(self, name: str, data: RecordsDataAlias, filter: Optional[Filter]) -> ActionResult:
        return await super().execute(name, data, filter)

    async def aggregate(
        self,
        filter: Optional[Filter],
        aggregation: Aggregation,
        limit: Optional[int] = None,
    ) -> List[AggregateResult]:
        with self.datasource.Session.begin() as session:  #  type: ignore
            dialect: Dialect = session.bind.dialect  #  type: ignore
            filter = cast(Filter, self._cast_filter(filter)) or None
            query = QueryFactory.build_aggregate(dialect, self, filter, aggregation, limit)
            res: List[Dict[str, Any]] = session.execute(query)  #  type: ignore
            return aggregations_to_records(res)

    @handle_sqlalchemy_error
    async def create(self, data: List[RecordsDataAlias]) -> List[RecordsDataAlias]:
        with self.datasource.Session.begin() as session:  #  type: ignore
            instances = QueryFactory.create(self, data)
            session.bulk_save_objects(instances, return_defaults=True)  # type: ignore
            return instances_to_records(self, instances)

    async def update(self, filter: Optional[Filter], patch: RecordsDataAlias) -> None:
        with self.datasource.Session.begin() as session:  #  type: ignore
            query = QueryFactory.update(self, filter, patch)
            session.execute(query)  # type: ignore

    def _cast_condition_tree(self, tree: ConditionTree) -> ConditionTree:
        if isinstance(tree, ConditionTreeLeaf):
            if TypeGetter.get(tree.value, None) == PrimitiveType.DATE:
                iso_format = tree.value
                if isinstance(iso_format, str):
                    if iso_format[-1] == "Z":
                        iso_format = iso_format[:-1]
                    iso_format = datetime.fromisoformat(iso_format).replace(
                        tzinfo=zoneinfo.ZoneInfo("UTC")
                    )  # type: ignore
                tree = tree.override({"value": iso_format})
        return tree

    def _cast_filter(self, filter: Union[Filter, PaginatedFilter, None]) -> Union[Filter, PaginatedFilter, None]:
        if filter and filter.condition_tree:
            filter = filter.override(
                {"condition_tree": filter.condition_tree.replace(self._cast_condition_tree)}  # type: ignore
            )
        return filter

    async def list(self, filter: PaginatedFilter, projection: Projection) -> List[RecordsDataAlias]:
        with self.datasource.Session.begin() as session:  #  type: ignore
            projection = self._normalize_projection(projection)
            filter = cast(PaginatedFilter, self._cast_filter(filter))
            query = QueryFactory.build_list(self, filter, projection)
            res = session.execute(query).all()  #  type: ignore
            records = projections_to_records(projection, res, filter.timezone)  # type: ignore
            return records

    async def delete(self, filter: Optional[Filter]) -> None:
        with self.datasource.Session.begin() as session:  #  type: ignore
            query = QueryFactory.delete(self, filter)
            session.execute(query)  # type: ignore

    async def get_form(
        self, name: str, data: Optional[RecordsDataAlias], filter: Optional[Filter]
    ) -> List[ActionField]:
        return await super().get_form(name, data, filter)


class SqlAlchemyDatasource(Datasource[SqlAlchemyCollection]):
    def __init__(self, Base: Any) -> None:
        super().__init__()
        self._base = Base
        self.Session = sessionmaker(self._base.metadata.bind)  # type: ignore
        self._create_collections()

    def build_mappers(self) -> Dict[str, Mapper]:
        mappers: Dict[str, Mapper] = {}
        for mapper in self._base.registry.mappers:
            mappers[mapper.persist_selectable.name] = mapper
        return mappers

    def _create_secondary_collection(self, table: Any):
        try:
            collection = self.get_collection(table.name)
        except DatasourceException:
            Secondary = type(str(table.name), tuple(), {})  # type: ignore
            mapper = self._base.registry.map_imperatively(Secondary, table)
            collection = SqlAlchemyCollection(table.name, self, table, mapper)
            self.add_collection(collection)
        return collection

    def _create_secondary_relation(
        self,
        collection: SqlAlchemyCollection,
        related_collection_name: str,
        many_to_many: ManyToMany,
    ):
        collection.add_field(
            related_collection_name.lower(),
            ManyToOne(
                foreign_collection=related_collection_name,
                foreign_key=many_to_many["origin_key"],
                foreign_key_target=many_to_many["origin_key_target"],
                type=FieldType.MANY_TO_ONE,
            ),
        )
        return many_to_many["foreign_collection"].lower()

    def _create_collections(self):
        mappers = self.build_mappers()
        for table in self._base.metadata.sorted_tables:
            if table.name in mappers:
                collection = SqlAlchemyCollection(table.name, self, table, mappers[table.name])
                self.add_collection(collection)
